#include "bool.h"
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/select.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>
#include <syslog.h>
#include <stdio.h>
#include "auth.h"
#include "util.h"
#include "child.h"
#include "network.h"
#include "socket.h"
#include "main.h"

static bool adminhandler(char *adminread);
static char *adminhandler_getauth(void);
static struct adminlist *adminhandler_getstate(void);
static struct adminlist *adminhandler_list(void);
static bool adminhandler_lock(char *user, char *authwrite, bool lock);
static bool authentication(char *authfile, char *user, char *pass);
static int getcommand(char **msg, bool admin);
static int getcommand_admincheck(char *buf) ;
static int getcommand_check(char *buf);
static bool getline(char **msg);
static bool inputhandler(char *authfile);
static char *pid2user(pid_t pid, struct adminlist *shmpt);
static pid_t user2pid(char *user, struct adminlist *shmpt);
static bool pid_not_in_list(struct adminlist *shmpt, pid_t pid);
/* static bool pipe_recv(struct pipe_msg_buf *msg); */
static bool pipe_recv_wait(struct pipe_msg_buf *msg);
static bool pipe_send(struct pipe_msg_buf msg);
static bool pipe_send_child(int command, char *user);
static void sendhelp(int auth);
static void sendhelp_admin(void);
static bool sendline(char *msg);
static bool sendnline(char *msg, int len);
static void send_parent_msg(int command, char *user);
static void getprint_constatus(void);
static void print_locked_users(char *authfile);
static bool sigsetup(void);

static char *peername;
static int sok;

int child(int chldsok, const int pwrite, const int pread, char *authfile)
{
	sok = chldsok;

	if (sigsetup()) /* ignore SIGPIPE */
		_exit(0);
	if (-1==dup2(pwrite, STDOUT_FILENO) || -1==dup2(pread, STDIN_FILENO)) {
		notice_err("dup2() failed");
		_exit(0);
	}
	peername = makepeername(sok); /* name of remote host */
	
	notice("%s: connected\n", peername);
	sendline("Hello\n");

	(void)inputhandler(authfile);
	notice("%s: closing\n", peername);
	_exit(0);
}

bool inputhandler(char *authfile)
{
	char *buf = NULL;
	char *user = NULL;
	bool connected = FALSE;
	int ret;
	int auth = 0; /* 0(noauth) 1(user given) 2(user/pass given) */

	ret = getcommand(&buf, FALSE);
	while(ret!=FATAL_TERM) {
		/* only COM_USER and COM_PASS allow arguments */
		if (ret==COM_USER) {
			if (!buf) {
				sendline("Invalid command:\n");
				sendhelp(auth); /* argument required */
			} else if (2==auth) {
				sendline("Invalid command:\n");
				sendhelp(auth);
				free(buf);
			} else {
				sendline("OKAY\n");
				user = buf;
				auth = 1; /* USER or PASS to follow */
			}
		} else if (ret==COM_PASS) {
			if (!buf) {
				/* requires argument */
				sendline("Invalid command:\n");
				sendhelp(auth);
			} else if (auth!=1) {
				/* needs to follow USER */
				sendline("Invalid command:\n");
				sendhelp(auth);
				free(buf);
			} else {
				/* sendline("OKAY\n");
				 * do we send OKAY because the command was
				 * valid, or do we not because the user/pass
				 * was invalid?
				 */
				if (authentication(authfile, user, buf)) {
					sendline("Invalid username/password\n");
					auth = 0; /* must USER again */
					memset(buf, 0, strlen(buf));
				} else {
					sendline("OKAY\n");
					auth = 2;
					memset(buf, 0, strlen(buf));
					free(buf);
					send_parent_msg(cAUTH, user);
					if (!strcmp(user, "admin")) {
						if (adminhandler(authfile)){
							return TRUE;
						}
						/* close */
						send_parent_msg(cDISCONNECT, NULL);
						auth = 0;
					}
				}
			}
		} else if (buf) {
			free(buf);
			sendline("Invalid command:\n");
			/* only USER/PASS may have argumemts */
			sendhelp(auth); 
		} else if (COM_HELP==ret) {
			sendline("OKAY\n");
			sendhelp(auth);
		} else if (FATAL_TIMEOUT==ret) {
			sendline("Timeout waiting for message\n");
			return TRUE;
		} else if (FATAL_SIGNAL==ret) {
			/* probably won't get to here */
			sendline("Killed\n");
			return TRUE;
		}else if (COM_QUIT==ret) {
			sendline("OKAY\n");
			return FALSE;
		} else if (COM_INVALID==ret) {
			sendline("Invalid command:\n");
			sendhelp(auth);
		} else if (COM_ALIVE==ret) {
			sendline("OKAY\n");
		} else if (auth<2) {
			sendline("Invalid command:\n");
			/* can only CONNECT/DISCONNECT/CLOSE if authed */
			sendhelp(auth); 
		} else if (COM_CONNECT==ret) {
			sendline("OKAY\n");
			if (!connected) {
				send_parent_msg(cCON, user);
				getprint_constatus();
			}
		} else if (COM_DISCONNECT==ret) {
			sendline("OKAY\n");
			/* if not already CONNECT'ed, server will handle it */
			send_parent_msg(cDISCONNECT, NULL);
		} else if (COM_CLOSE==ret) {
			sendline("OKAY\n");
			send_parent_msg(cDISCONNECT, NULL);
			auth = 0;
		} else {
			notice("%s: bug in inputhandler()\n", peername);
		}
		ret = getcommand(&buf, FALSE);
	} /* while(ret!=FATAL_TERM) */
	return TRUE;
}

void sendhelp(int auth)
{
	switch(auth) {
		case 0: sendline("Enter USER username, HELP, QUIT\n");
			break;
		case 1: sendline("Enter USER username, PASS password, HELP, "
					"QUIT\n");
			break;
		default: sendline("Enter HELP, CONNECT, DISCONNECT, ALIVE, "
					 "QUIT, CLOSE\n");
			 break;
	}
	return;
}

void sendhelp_admin(void)
{
	sendline("Enter SHOWLOCKS SHUTDOWN, RESTART, ALIVE, CLOSE, HELP, "
			"LIST, QUIT\nEnter KILL, LOCK, UNLOCK, ZAP with an "
			"argument of either pid or %username\n");
	return;
}

/* gets input from user, parses for commands */
int getcommand(char **msg, bool admin)
{
	char *buf;
	int com_len;
	fd_set sdfd;
	struct timeval tv;
	int ret;

	/* must receive data before select times out */
	if (set.keep_alive_timeout> 0) {
		memset(&sdfd, 0, sizeof(sdfd));
		memset(&tv, 0, sizeof(tv));
		FD_SET(sok, &sdfd);
		tv.tv_sec = set.keep_alive_timeout;

		if (1!=select(sok+1, &sdfd, NULL, NULL, &tv)) {
			*msg = NULL;
			if (EINTR==errno) {
				return FATAL_SIGNAL;
			}
			return FATAL_TIMEOUT;
		}
	}
	
	if (getline(&buf)) {
		return FATAL_TERM;
	}
	if ((com_len = strchr(buf, ' ') - buf) > 0) { 
		/* argument */
		if (alloc_copy(msg, buf+com_len+1)) 
			return FATAL_TERM;
		/* command */
		*(buf+com_len) = '\0';
	} else {
		/* command without an argument */
		*msg = NULL;
	}
	if (admin) {
		ret = getcommand_admincheck(buf);
	} else {
		ret = getcommand_check(buf);
	}
	free(buf);
	return ret;
}

int getcommand_check(char *buf)
{
	if (!strcasecmp(buf, "HELP")) {
		return COM_HELP;
	} else if (!strcasecmp(buf, "CONNECT")) {
		return COM_CONNECT;
	} else if (!strcasecmp(buf, "DISCONNECT")) {
		return COM_DISCONNECT;
	} else if (!strcasecmp(buf, "ALIVE")) {
		return COM_ALIVE;
	} else if (!strcasecmp(buf, "USER")) {
		return COM_USER;
	} else if (!strcasecmp(buf, "PASS")) {
		return COM_PASS;
	} else if (!strcasecmp(buf, "QUIT")) {
		return COM_QUIT;
	} else if (!strcasecmp(buf, "CLOSE")) {
		return COM_CLOSE;
	} else {
		return COM_INVALID;
	}
}

int getcommand_admincheck(char *buf) 
{
	if (!strcasecmp(buf, "SHUTDOWN")) {
		return ADM_SHUTDOWN;
	} else if (!strcasecmp(buf, "RESTART")) {
		return ADM_RESTART;
	} else if (!strcasecmp(buf, "KILL")) {
		return ADM_KILL;
	} else if (!strcasecmp(buf, "LOCK")) {
		return ADM_LOCK;
	} else if (!strcasecmp(buf, "UNLOCK")) {
		return ADM_UNLOCK;
	} else if (!strcasecmp(buf, "SHOWLOCKS")) {
		return ADM_SHOWLOCKS;
	} else if (!strcasecmp(buf, "ZAP")) {
		return ADM_ZAP;
	} else if (!strcasecmp(buf, "CLOSE")) {
		return ADM_CLOSE;
	} else if (!strcasecmp(buf, "QUIT")) {
		return ADM_QUIT;
	} else if (!strcasecmp(buf, "ALIVE")) {
		return ADM_ALIVE;
	} else if (!strcasecmp(buf, "HELP")) {
		return ADM_HELP;
	} else if (!strcasecmp(buf, "LIST")) {
		return ADM_LIST;
	} else if (!strcasecmp(buf, "RECONNECT")) {
		return ADM_RECONNECT;
	} else {
		return ADM_INVALID;
	}
}

/* send pipe message
 * username only really needed with cCON */
bool pipe_send_child(int command, char *user)
{
	struct pipe_msg_buf msg;

	msg.from = getpid();
	msg.command = command;
	if (!user) {
		memset(msg.username, 0, USER_LEN);
	} else {
		strncpy(msg.username, user, USER_LEN);
		msg.username[USER_LEN]='\0';
	}
	if (pipe_send(msg)) {
		notice("%s: parent failed to respond\n", peername);
		return TRUE; /* quite possibly stuffed now */
	}
	return FALSE;
}

void send_parent_msg(int command, char *user)
{
	(void)pipe_send_child(command, user);
	return;
}

/* get/check username/password from peer */
bool authentication(char *authfile, char *user, char *pass)
{
	if (check_pass(pass, inpass(user, authfile, ':'))) {
		notice("%s: invalid auth\n", peername);
		return TRUE;
	}
	notice("%s: %s authenticated\n", peername, user);
	return FALSE;
}

bool adminhandler(char *authread)
{
	int our_pid;
	char *buf;
	struct adminlist *shmpt;
	char *authwrite;
	char *user;
	pid_t victim;
	int ret;
	int exit_time = Q_NORMAL;

	our_pid = getpid();
	authwrite = NULL;
	user = NULL;

	if (!(shmpt = adminhandler_getstate())) {
		return TRUE;
	}

	ret = getcommand(&buf, TRUE);
	while(ret!=FATAL_TERM) {
		if (ret==ADM_INVALID) {
			sendline("Invalid command:\n");
			sendhelp_admin();
		} else if (ret==FATAL_TIMEOUT) {
			sendline("Timeout waiting for message\n");
			exit_time = Q_TRUE;
		} else if (ret==FATAL_SIGNAL) {
			sendline("Killed\n");
			exit_time = Q_TRUE;
		}else if (!buf) {
			/* no arguments */
			if (ret==ADM_SHUTDOWN) {
				sendline("OKAY\n");
				if (-1==kill(getppid(), SIGHUP)) {
					notice("%s: shutting down failed\n", 
							peername);
				} else {
					exit_time = Q_TRUE;
				}
			} else if (ret==ADM_RESTART) {
				sendline("OKAY\n");
				send_parent_msg(cADMIN_RESTART, NULL);
				exit_time = Q_TRUE;
			} else if (ret==ADM_RECONNECT) {
				sendline("OKAY\n");
				if (-1==kill(getppid(), SIGUSR2)) {
					notice("%s: reconnecting\n", peername);
				}
			} else if (ret==ADM_CLOSE) {
				sendline("OKAY\n");
				exit_time = Q_FALSE;
			} else if (ret==ADM_QUIT) {
				sendline("OKAY\n");
				exit_time = Q_TRUE;
			} else if (ret==ADM_ALIVE) {
				sendline("OKAY\n");
			} else if (ret==ADM_HELP) {
				sendline("OKAY\n");
				sendhelp_admin();
			} else if (ret==ADM_LIST) {
				sendline("OKAY\n");
				shmdt(shmpt);
				if (!(shmpt = adminhandler_list())) {
					exit_time = Q_TRUE;
				}
			} else if (ret==ADM_SHOWLOCKS) {
				sendline("OKAY\n");
				print_locked_users(authread);
				sendline("\n");
			} else {
				sendline("Invalid command:\n");
				sendhelp_admin();
			}
		/* there is an argument */
		} else {
			/* argument is pid or user if starts with % */
			if (*buf=='%') {
				if (alloc_copy(&user, buf+1)){
					exit_time = Q_TRUE;
					user = NULL;
					victim = 0;
				} else {
					victim = user2pid(user, shmpt);
				}
				/* victim = 0 if user not authed currently */
			} else {
				victim = atoi(buf);
				user = pid2user(victim, shmpt); /* or NULL */
			}
			free(buf);
			
			if (pid_not_in_list(shmpt, victim)) {
				sendline("Invalid PID\n");
			} else if (ret==ADM_KILL) {
				if (victim > 0) {
					sendline("OKAY\n");
					if (-1==kill(victim, SIGHUP)) {
						sendline("KILL failed\n");
					}
				} else {
					sendline("Invalid PID or USER\n");
				}
			} else if (ret==ADM_LOCK) {
				sendline("OKAY\n");
				if (!authwrite) {
					authwrite = adminhandler_getauth();
				}
				adminhandler_lock(user, authwrite, TRUE);
			} else if (ret==ADM_UNLOCK) {
				sendline("OKAY\n");
				if (!authwrite) {
					authwrite = adminhandler_getauth();
				}
				adminhandler_lock(user, authwrite, FALSE);
			} else if (ret==ADM_ZAP) {
				sendline("OKAY\n");
				if (victim > 0) {
					if (-1==kill(victim, SIGHUP)) {
						sendline("KILL failed\n");
					}
				}
				if (!authwrite) {
					authwrite = adminhandler_getauth();
				}
				adminhandler_lock(user, authwrite, TRUE);
			} else {
				notice("%s: bug in adminhandler()\n", peername);
			}
		}
		if (exit_time==0) {
			ret = getcommand(&buf, TRUE);
		} else {
			ret = FATAL_TERM;
		}
	} /* while(ret!=FATAL_TERM) */
	shmdt(shmpt);
	if (NULL!=authwrite) {
		shmdt(authwrite);
	}
	if (exit_time==Q_FALSE) {
		return FALSE;
	}
	return TRUE;
}

char *pid2user(pid_t pid, struct adminlist *shmpt)
{
	while (shmpt->pid!=0) {
		if (shmpt->pid==pid) {
			return shmpt->username;
		}
		shmpt++;
	}
	return NULL;
}

pid_t user2pid(char *user, struct adminlist *shmpt)
{
	while(shmpt->pid!=0) {
		if (!strcmp(user, shmpt->username)) {
			return shmpt->pid;
		}
		shmpt++;
	}
	return 0;
}
	
bool adminhandler_lock(char *user, char *authwrite, bool lock)
{
	if (!user) {
		sendline("account %USER not found\n");
		return TRUE;
	}
	if (lock) {
		if (!(authwrite = inpass(user, authwrite, ':'))) {
			sendline("enabled account %USER not found\n");
			return TRUE;
		}
		*(authwrite-1) = '!'; /* overwrite : */
	} else {
		if (!(authwrite = inpass(user, authwrite, '!'))) {
			sendline("disabled account %USER not found\n");
			return TRUE;
		}
		*(authwrite-1) = ':'; /* overwrite ! */
	}
	return FALSE;
}

char *adminhandler_getauth(void)
{
	struct pipe_msg_buf msg;
	char *authwrite;

	send_parent_msg(cADMIN_AUTHWRITE, NULL);
	if (pipe_recv_wait(&msg)) {
		return NULL;
	}

	if (!(authwrite = shmat(msg.command, 0, 0))) {
		notice("%s: couldn't attached shared mem for admin\n", peername);
		return NULL;
	}
	return authwrite;
}
	
struct adminlist *adminhandler_list(void)
{
	struct adminlist *shmpt;
	struct adminlist *tmp;
	char *buf;
	pid_t our_pid;

	if (!(buf = (char *)malloc(256 * sizeof(char)))) {
		notice_err("malloc() failed");
		return NULL;
	}
	our_pid = getpid();
	if (!(shmpt = adminhandler_getstate())) {
		return NULL;
	}
	tmp = shmpt;
	/* tmp->pid==0 is terminator */
	while (tmp->pid!=0) {
		if (tmp->pid!=our_pid) {
			/* portability warning */
			snprintf(buf, 255, "%d %s %s %s\n", 
					tmp->pid, 
					tmp->peername, 
					tmp->username, 
					tmp->connected==TRUE?
					"CON":"!CON");
			sendline(buf);
			free(buf);
		}
		++tmp;
	}
	return shmpt;
}

struct adminlist *adminhandler_getstate(void)
{
	struct pipe_msg_buf msg;
	struct adminlist *shmpt;

	memset(&msg, 0, sizeof(msg));

	send_parent_msg(cADMIN_LIST, "admin");
	do {
		if (pipe_recv_wait(&msg)) {
			return NULL;
		}
		
		if (-1==msg.command) {
			sendline("Admin already connected\n");
			notice("%s: admin already connected\n", peername);
			return NULL;
		}
	} while (msg.from==1); /* ignore messages from parent :-) */

	/* the msg.command is shmid as set by parent */
	if (!(shmpt = shmat(msg.command, 0, 0))) {
		notice("%s: couldn't attach shared mem for admin\n", peername);
		return NULL;
	}
	/* admin parent can now delete its shmem reference and use shmctl so
	 * reference will be deleted when we release */
	notice_debug("adminhandler_getstate sending SIGUSR1 to %d", msg.from);
	kill(msg.from, SIGUSR1);

	return shmpt;
}

void print_locked_users(char *authfile)
{
	size_t i;
	
	while ((i = substringlen(authfile, '!'))) {
		sendnline(authfile, i+1); /* including ! */
		authfile = strchr(authfile, '\n');
		++authfile;
	}
	return;
}
	
bool pid_not_in_list(struct adminlist *shmpt, pid_t pid)
{
	if (!pid)
		return FALSE; /* using %user instead */ 
	while (shmpt->pid!=0) {
		if (shmpt->pid==pid)
			return FALSE;
		shmpt++;
	}
	return TRUE;
}

bool sendline(char *msg)
{
	struct timeval tv;
	tv.tv_sec = 10;
	tv.tv_usec = 0;
	switch(sok_send(sok, &tv, msg, strlen(msg))) {
		case SOK_SUCCESS: break;
		case SOK_FAILED : notice("%s: sendline() failed\n", peername);
				  return TRUE;
		case SOK_TIMEOUT: notice("%s: sendline() timeout\n", peername);
				  return TRUE;
	}
	return FALSE;
}

bool sendnline(char *msg, int len)
{
	struct timeval tv;
	tv.tv_sec = 10;
	tv.tv_usec = 0;
	switch(sok_send(sok, &tv, msg, len)) {
		case SOK_SUCCESS: break;
		case SOK_FAILED : notice("%s: sendline() failed\n", peername);
				  return TRUE;
		case SOK_TIMEOUT: notice("%s: sendline() timeout\n", peername);
				  return TRUE;
	}
	return FALSE;
}

bool getline(char **msg)
{
	struct timeval tv;
	struct timeval *tvpt;
	if (set.keep_alive_timeout > 0) {
		tv.tv_sec = set.keep_alive_timeout;
		tv.tv_usec = 0;
		tvpt = &tv;
	} else {
		tvpt = NULL;
	}
	if (!(*msg = malloc(sizeof(char) * 256))) {
		notice_err("malloc() failed");
		return TRUE;
	}
	switch(sok_getstring(*msg, 255, sok, tvpt)) {
		case SOK_SUCCESS: break;
		case SOK_FAILED : /* notice("%s: getline() failed\n", peername);
				     */
				  notice("%s: connection reset by peer\n", 
						  peername);
				  return TRUE;
		case SOK_TIMEOUT: /*notice("%s: getline() timeout\n", peername);
				    */
				  notice("%s: peer failed to respond\n",
						  peername);
				  return TRUE;
		case SOK_TOOLONG: /* notice("%s: getline() input error\n",
						  peername); */
				  notice("%s: garbage input from peer\n",
						  peername);
				  break;
	}
	return FALSE;
}

bool pipe_send(struct pipe_msg_buf msg)
{
	struct timeval tv;
	tv.tv_sec = 5;
	tv.tv_usec = 0;
	switch(sok_send(STDOUT_FILENO, &tv, &msg, sizeof(msg))) {
		case SOK_SUCCESS: break;
		case SOK_FAILED : notice("%s: pipe_send() failed\n", peername);
				  return TRUE;
		case SOK_TIMEOUT: notice("%s: pipe_send() timeout\n", peername);
				  return TRUE;
	}
	return FALSE;
}

#if 0
/* everything uses pipe_recv_wait so far */
bool pipe_recv(struct pipe_msg_buf *msg)
{
	struct timeval tv;
	tv.tv_sec = 5;
	tv.tv_usec = 0;
	switch(sok_recv(STDIN_FILENO, &tv, msg, sizeof(msg))) {
		case SOK_SUCCESS: break;
		case SOK_FAILED : notice("%s: pipe_recv() failed\n", peername);
				  return TRUE;
		case SOK_TIMEOUT: notice("%s: pipe_recv() timeout\n", peername);
				  return TRUE;
	}
	return FALSE;
}
#endif

bool pipe_recv_wait(struct pipe_msg_buf *msg)
{
	switch(sok_recv(STDIN_FILENO, NULL, msg, sizeof(struct pipe_msg_buf))) {
		case SOK_SUCCESS: break;
		case SOK_FAILED : notice("%s: pipe_recv() failed\n", peername);
				  return TRUE;
		case SOK_TIMEOUT: notice("%s: pipe_recv() timeout\n", peername);
				  return TRUE;
	}
	return FALSE;
}

void getprint_constatus(void)
{
	struct pipe_msg_buf msg;
	if (pipe_recv_wait(&msg))
		return;
	switch(msg.command) {
		case CONS_DISCONNECTED: break; /* parent will send */
		case CONS_CONNECTING  : sendline("#Connecting\n");
					break;
		case CONS_CONNECTED   : sendline("#Connected\n");
					break;
	}
	return;
}

bool sigsetup(void)
{
	struct sigaction cursig;

	memset(&cursig, 0, sizeof(cursig));
	cursig.sa_handler = SIG_IGN;
	if (-1==sigaction(SIGPIPE, &cursig, NULL)) {
		notice_err("%s: sigaction failed", peername);
		return TRUE;
	}
	return FALSE;
}
