#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/wait.h>
#include <stdio.h>
#include "shared.h"
#include "util.h"
#include "exec.h"
#include "main.h"

static bool make_exec_command(char **buf, const char *fmt, 
		const char *peername, const char *username, int maxlen);
static bool on_connect_execute(const char *command);
static bool quad_out_of_range(unsigned long peer[4], unsigned long host[4], 
		unsigned long mask[4]);
static void host_to_quad(const char *host, unsigned long *quad);

bool on_connect(bool connect, const char *peername, const char *username)
{
	char *command;

	if (!strcmp(username, "%unknown")) {
		return TRUE;  /* better safe than sorry */
	}

	/* we rely on item_add rejecting invalid IP's */
	if (connect) {
		if (make_exec_command(&command, set.commandon, peername,
					username, 1024))
			return TRUE;
	} else {
		if (make_exec_command(&command, set.commandoff,peername,
					username,1024)) 
			return TRUE;
	}

	if (on_connect_execute(command)) {
		free(command);
		return TRUE;
	}
	free(command);
	return FALSE;
}

bool on_connect_execute(const char *command)
{
	pid_t pid;

	switch (pid = fork()) {
		case -1 : notice_err("fork() failed");
			  return TRUE;
		case  0 : if (-1==execl("/bin/sh", "sh", "-c", command, NULL)) {
				  notice_err("running %s failed", command);
			  }
			  _exit(0);
		default : break; /* ignoring exit status of command */
	}
	return FALSE;
}
			  
extern bool ip_out_of_range(const char *peername)
{
	char *allowed_host;
	char *allowed_mask;
	int len;
	unsigned long peernamequad[4];
	unsigned long hostquad[4];
	unsigned long maskquad[4];

	if (!peername) {
		return TRUE;
	}
	if (-1==inet_addr(peername)) {
		return TRUE; /* invalid address */
	}
	if (!set.allowed_hostmask) {
		return FALSE; /* assume 0.0.0.0/0.0.0.0, must be okay */
	}

	if (!(allowed_mask = strchr(set.allowed_hostmask, '/')))
		return TRUE;
	if (alloc_copy(&allowed_mask, allowed_mask+1))
		return TRUE;
	len = strlen(set.allowed_hostmask) - strlen(allowed_mask) - 1;
	allowed_host = (char *)malloc(sizeof(char)*len);
	if (!allowed_host) {
		free(allowed_mask);
		return TRUE;
	}
	strncpy(allowed_host, set.allowed_hostmask, len);
	
	host_to_quad(peername, peernamequad);
	host_to_quad(allowed_mask, maskquad);
	host_to_quad(allowed_host, hostquad);

	free(allowed_mask);
	free(allowed_host);

	return quad_out_of_range(peernamequad, hostquad, maskquad);
}
			
/* FIXME: I've had a segfault on strchr here before, but I put some debugging
 * in and since then haven't been able to reproduce it. It may relate to
 * set.allow_hostmask being corrupted */
void host_to_quad(const char *host, unsigned long *quad)
{
        const char *s;
        unsigned long *pt;

        for (pt=quad, s=host-1; s; s=(char *)strchr(s, '.')) {
                *pt++ = strtoul(++s, (char **)NULL, 10);
        }
        return;
}
	
bool quad_out_of_range(unsigned long peer[4], unsigned long host[4], 
		unsigned long mask[4])
{
        int i;

        for (i=0; i<4; i++) {
		if ((peer[i] & mask[i]) != (host[i] & mask[i])) {
			return TRUE;
		}
	}
	return FALSE;
}

bool make_exec_command(char **buf, const char *fmt, const char *peername, 
		const char *username, int maxlen)
{
	char *str;
	const char *s;
	int i;
	int len;

	if (!fmt || !peername || !username) {
		return TRUE;
	}
	if (!(*buf = (char *)malloc(maxlen))) {
		return TRUE;
	}
	for (str=*buf; *fmt && maxlen > 0; ++fmt) {
		if (*fmt!='%') {
			*str++ = *fmt;
			--maxlen;
			continue;
		}

		++fmt;
		switch(*fmt) {
			case 'b' : s = peername;
				   break;
			case 'u' : s = username;
				   break;
			default  : continue;
		}
		len = strlen(s);
		for (i=0; i<len && maxlen > 0; i++) {
			*str++=*s++;
			maxlen--;
		}
	}
	if (maxlen < 0) {
		free(*buf);
		return TRUE;
	}
	*str = '\0';
	return FALSE;
}
