#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include "util.h"
#include "network.h"
#include "socket.h"
#include "exec.h"
#include "item.h"
#include "main.h"

static struct clients *item_from_pid(struct clients *start, pid_t pid);

/* removes the item with specified pid */
int item_rem(struct clients *start, struct clients **end, pid_t pid)
{
	struct clients *tmp;	/* holds pointer while we free item */

	start = item_from_pid(start, pid);
	if (!start) {
		return ITEM_NOTFOUND;
	}
	if (start->next==*end) {
		*end = start; /* removing the last item is special */
	}
	tmp = start->next->next;
	(void)close(start->next->sok);
	(void)close(start->next->pwrite);
	if (start->next->connected) {
		notice_debug("connected child exited without disconnect");
		(void)on_connect(FALSE, start->next->peername, 
				 start->next->username);
		free(start->next); 
		start->next = tmp;
		return ITEM_DISCONNECT;
	}
	free(start->next); 
	start->next = tmp;
	return ITEM_NORMAL;
}

/* adds an item */
bool item_add(struct clients **end, pid_t pid, int sok, int pwrite, 
		char *peername) 
{
	(*end)->next = (struct clients *)malloc(sizeof(struct clients));
	if (!(*end)->next) {
		notice("malloc() failed\n");
		if (-1==kill(pid, SIGHUP)) {
			notice("kill() failed\n");
		}
		return TRUE;
	}
	(*end)->next->pid = pid;
	(*end)->next->sok = sok;
	(*end)->next->connected = FALSE;
	(*end)->next->pwrite = pwrite;
	strncpy((*end)->next->username, "%unknown", USER_LEN);
	(*end)->next->peername = peername;
	*end = (*end)->next;
	(*end)->next = NULL;
	return FALSE;
}

/* kills all processes in list and removes records */
void killall(struct clients *start, struct clients **end, pid_t adminchildpid)
{
	struct clients *tmp;
	struct clients *pt;

	if (adminchildpid > 0) {
		(void)kill(adminchildpid, SIGHUP);
	}

	pt = start->next;
	while (pt) {
		(void)kill(pt->pid, SIGHUP);
		(void)close(pt->sok);
		(void)close(pt->pwrite);
		(void)waitpid(pt->pid, NULL, 0);
		tmp = pt;
		pt = pt->next;
		free(tmp);
	}
	start->next = NULL;
	*end = start;
	return;
}

/* set connected flag in list for specified pid */
bool item_con(pid_t pid, struct clients *start, bool action)
{
	start = item_from_pid(start, pid);
	if (!start) {
		notice_debug("pid not found");
		return TRUE;
	}
	if (!action==start->next->connected) {
		start->next->connected = action;
		(void)on_connect(TRUE, start->next->peername, 
				 start->next->username);
		return FALSE;
	} else {
		return TRUE;
	}
}

/* set authenticated info (via username setting)
if username==NULL, unauthenticated else authenticated */
bool item_auth(pid_t pid, char *username, struct clients *start, bool action)
{
	start = item_from_pid(start, pid);
	if (!start->next) {
		notice("pid not found\n");
		return TRUE;
	}
	if (action && '%'==start->next->username[0]) {
		strncpy(start->next->username, username, USER_LEN); /* auth */
	} else if (!action && '%'!=start->next->username[0]) {
		start->next->username[0] = '%'; /* deauth */
	}
	return FALSE;
}

/* returns a pointer to the item where: item->next contains pid */
struct clients *item_from_pid(struct clients *start, pid_t pid)
{
	if (!start->next) {
		return (struct clients *)NULL;
	}
	while (start->next->pid!=pid) {
		start = start->next;
		if (!start->next) {
			return (struct clients*)NULL;
		}
	}
	return start;
}

/* returns the pipe the parent can write to for the specified pid
 * sok_from_pid is different and is in local.c (uses start->next) */
int pipe_from_pid(pid_t pid, struct clients *start)
{
	if (!start->next) {
		return -1;
	}
	while (start->next->pid!=pid) {
		start = start->next;
		if (!start->next) {
			return -1;
		}
	}
	return start->next->pwrite;
}

void setall_disconnected(struct clients *start)
{
	if (!start) {
		return;
	}
	for (start=start->next; start; start=start->next) {
		start->connected = FALSE;
		(void)on_connect(FALSE, start->peername, start->username);
	}
	return;
}
