#include <syslog.h>
#include <limits.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#ifdef __GNU_LIBRARY__
#include <getopt.h>
#endif
#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include "rets.h"
#include "network.h"
#include "util.h"
#include "rcfile.h"
#include "main.h"

static void helpinfo(char *progname);

int main(int argc, char *argv[]) 
{
	LOG_TO_SYSLOG = FALSE;

	if (setargs(argc, argv)) {
		return EXIT_FAILURE;
	}
	while (server()==QRESTART);
	notice("exiting\n");
	return EXIT_SUCCESS;
}

bool setargs(int argc, char *argv[])
{
	char ret;
	char *rcfile;
	char **pt;

	/* initialise error space */
	if (!(error_messages = (char *)malloc(MAX_LOG_LENGTH)))
		return TRUE;
	/* defaults */
	set.debug = FALSE;
	set.keep_alive_timeout = 20;
	set.port = 8000;
	set.bind_ip = inet_addr("0.0.0.0");
	set.maxcons = 10;
	if (alloc_copy(&set.authfile, "/etc/dwunauth"))
		return TRUE;
	if (alloc_copy(&set.dwunpid, "/var/run/dwun.pid")) 
		return TRUE;
	set.commandon = NULL;
	set.commandoff = NULL;
	set.allowed_hostmask = NULL;
	if (alloc_copy(&rcfile, "/etc/dwunrc"))
		return TRUE;

	/* options from rcfile */
	for (pt = argv; *pt; ++pt) {
		if (**pt=='-' && *(*pt+1)=='r') {
			if (rcfile)
				free(rcfile);
			if (alloc_copy(&rcfile, *(pt+1)))
				return TRUE;
			break;
		}
	}
	if (get_rcfile_settings(rcfile)) {
		return TRUE;
	}

	/* command-line settings override defaults and those from rcfile */
	while ((ret = getopt(argc, argv, "p:b:m:a:k:r:dvh"))!=EOF) {
		switch(ret) {
			case 'p' : /* port to bind to */
				set.port =(short)strtoul(optarg,(char **)NULL, 
						10);
			        if (set.port > USHRT_MAX || 0==set.port) {
					notice("Invalid port\n");
					return TRUE;
				}
				break;
			case 'b' :
				set.bind_ip = inet_addr(optarg);
				break;
			case 'm' : /* max clients allowed */
				set.maxcons = atoi(optarg);
				break;
			case 'a' :
				if (set.authfile)
					free(set.authfile);
				if (alloc_copy(&set.authfile, optarg))
					return TRUE;
				break;
			case 'k' :
				set.keep_alive_timeout = atoi(optarg);
				break;
			case 'r' : break; /* already dealt with */
			case 'd' :
				set.debug = TRUE;
				break;
			case 'v' : 
				notice("DWUN version %s\n", DWUN_VERSION);
				notice("Copyright (C) 1999 Tim Sutherland\n");
				notice("See the file COPYRIGHT in the distribution\n");
				exit(EXIT_SUCCESS);
				break;
			case 'h' : /* help */
				helpinfo(argv[0]);
				exit(EXIT_SUCCESS);
			case ':' :
				notice("option requires an argument\n");
				helpinfo(argv[0]);
				return TRUE;
			case '?' :
				notice("unknown option\n");
				helpinfo(argv[0]);
				return TRUE;
			default : notice_debug("bug in setargs");
				  break;
		}
	}
	return FALSE;
}

void helpinfo(char *progname)
{
	notice("DWUN version %s\n", DWUN_VERSION);
	notice("Copyright (C) 1999 Tim Sutherland\n");
	notice("See the file COPYRIGHT in the distribution\n\n");
	notice("Usage: %s [OPTION] ...\n\n", progname);
	notice("-p\tport\n");
	notice("-b\tIP to bind to\n");
	notice("-m\tmaximum clients allowed\n");
	notice("-a\tauthentication file\n");
	notice("-k\tseconds to wait for client keep_alive\n");
	notice("-r\tread options from file\n");
	notice("-d\tenable debugging mode\n");
	notice("-v\tversion\n");
	notice("-h\tthis help\n");
	return;
}
