#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "util.h"
#include "shared.h"
#include "rcfile.h"
#include "main.h"

static bool mksettings(FILE *fp);
static int getsetting(FILE *fp, char **argument);
static int getsetting_check(char *buf);
static void setstring(char **s, char *msg);

bool get_rcfile_settings(char *filename)
{
	FILE *fp;

	if (!filename) {
		return FALSE; /* no rcfile */
	}
	if (!(fp = fopen(filename, "r"))) {
		notice_err("cannot open %s", filename);
		return TRUE;
	}
	if (mksettings(fp)) {
		fclose(fp);
		return TRUE;
	}
	fclose(fp);
	return FALSE;
}
	
bool mksettings(FILE *fp)
{
	int ret;
	char *arg;
	int line = 0;
	bool error = FALSE;
	
	if (RC_FATAL==(ret = getsetting(fp, &arg)))
		return TRUE;
	while (ret!=RC_EOF) {
		++line;
		if (ret==RC_INVALID) {
			notice("error in rc file line %d\n", line);
			error = TRUE;
		} else if (!arg) {
			switch(ret) {
				case RC_COMMENT : break;
				case RC_DEBUG   : set.debug = TRUE;
						  break;
				default		: notice("error in rc file "
							 "line %d\n", line);
						  break;
			}
		} else {
			switch(ret) {
				case RC_PORT	: set.port = atoi(arg);
						  break;
				case RC_BIND	: set.bind_ip = inet_addr(arg);
						  break;
				case RC_MAXCON	: set.maxcons = atoi(arg);
						  break;
				case RC_KEEPALIVE : 
						  set.keep_alive_timeout =
							  atoi(arg);
						  break;
				case RC_ONCONNECT : 
						  setstring(&set.commandon,arg);
						  break;
				case RC_ONDISCONNECT : 
						  setstring(&set.commandoff,
								  arg);
						  break;
				case RC_HOSTMASK : 
						  setstring(&set.allowed_hostmask, arg);
						  break;
				case RC_AUTHFILE :
						  setstring(&set.authfile, arg);
						  break;
				case RC_PIDFILE  : 
						  setstring(&set.dwunpid, arg); 
						  break;
				default		 :
						  notice("error in rc file "
							  "line %d\n", line);
						  break;
			}
		}
		if (RC_FATAL==(ret = getsetting(fp, &arg)))
			return TRUE;
	}
	if (error)
		return TRUE; /* having printed all other errors */
	return FALSE;
}

void setstring(char **s, char *msg)
{
	if (*s)
		free(*s);
	*s = msg;
	return;
}

int getsetting(FILE *fp, char **argument)
{
	char *buf;
	int com_len;
	int ret;
	char *pt;

	if (!(buf = (char *)malloc(256)))
		return RC_FATAL;
	if (!fgets(buf, 256, fp))
		return RC_EOF;
	if ((com_len = strchr(buf, ' ') - buf) > 0) {
	 	/* argument */
		if (alloc_copy(argument, buf+com_len+1))
                        return RC_FATAL;
		for (pt=*argument; *pt; ++pt);
		*(pt-1) = '\0'; /* bye bye \n */
                /* command */
                *(buf+com_len) = '\0';
        } else {
                /* command without an argument */
                *argument = NULL;
		for (pt=buf; *pt; ++pt);
		*(pt-1) = '\0'; /* bye bye \n */
        }
	if (*buf=='#' || *buf=='\0') {
		free(buf);
		*argument = NULL;
		return RC_COMMENT; /* starts with # or blank line */
	}
	ret = getsetting_check(buf);
	free(buf);
	return ret;
}

int getsetting_check(char *buf)
{
	if (!strcasecmp(buf, "port")) {
		return RC_PORT;
	} else if (!strcasecmp(buf, "bind")) {
		return RC_BIND;
	} else if (!strcasecmp(buf, "maxcon")) {
		return RC_MAXCON;
	} else if (!strcasecmp(buf, "debug")) {
		return RC_DEBUG;
	} else if (!strcasecmp(buf, "keepalive")) {
		return RC_KEEPALIVE;
	} else if (!strcasecmp(buf, "onconnect")) {
		return RC_ONCONNECT;
	} else if (!strcasecmp(buf, "ondisconnect")) {
		return RC_ONDISCONNECT;
	} else if (!strcasecmp(buf, "hostmask")) {
		return RC_HOSTMASK;
	} else if (!strcasecmp(buf, "authfile")) {
		return RC_AUTHFILE;
	} else if (!strcasecmp(buf, "pidfile")) {
		return RC_PIDFILE;
	} else {
		return RC_INVALID;
	}
}
