#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include "util.h"
#include "socket.h"
#include "main.h"

int sok_send(int sok, struct timeval *tv, void *msg, size_t len_left)
{
	fd_set sokset;
	struct timeval tvbak;
	int ret;

	FD_ZERO(&sokset);
	FD_SET(sok, &sokset);

	while (len_left > 0) {
		do {
			ret = write(sok, msg, len_left);
		} while (ret==-1 && errno==EINTR);
		if (ret==-1) {
			if (errno!=EAGAIN)
				return SOK_FAILED;
			if (tv!=NULL)
				memcpy(tv, &tvbak, sizeof(*tv));
			do {
				ret = select(sok+1, NULL, &sokset, NULL, tv);
			} while (ret==-1 && errno==EINTR);
			switch(ret) {
				case -1: return SOK_FAILED;
				case  0: return SOK_TIMEOUT;
				default: break;
			}
		} else
			len_left -= ret;
	}
	return SOK_SUCCESS;
}

int sok_recv(int sok, struct timeval *tv, void *msg, size_t len_left)
{
	fd_set sokset;
	struct timeval tvbak;
	int ret;

	FD_ZERO(&sokset);
	FD_SET(sok, &sokset);

	while (len_left > 0) {
		do {
			ret = read(sok, msg, len_left);
		} while (ret==-1 && errno==EINTR);
		if (ret==-1) {
			if (errno!=EAGAIN)
				return SOK_FAILED;
			if (tv!=NULL)
				memcpy(tv, &tvbak, sizeof(*tv));
			do {
				ret = select(sok+1, &sokset, NULL, NULL, tv);
			} while (ret==-1 && errno==EINTR);
			switch(ret) {
				case -1: return SOK_FAILED;
				case  0: return SOK_TIMEOUT;
				default: break;
			}
		} else if (ret==0) {
			return SOK_FAILED; /* EOF */
		} else  {
			len_left -= ret;
		}
	}
	return SOK_SUCCESS;
}

int sok_getstring(char *msg, int maxline, int sok, struct timeval *tv)
{
	int numchars;
	int ret;
	struct timeval tvbak;
	fd_set reading;

	if (tv!=NULL) {
		memcpy(&tvbak, tv, sizeof(tvbak));
	}
	numchars = 0;
	while (numchars < maxline) {
		do {
			ret = read(sok, msg+numchars, 1);
		} while (ret==-1 && errno==EINTR);
		if (ret==-1) {
			if (errno!=EAGAIN) {
				free(msg);
				return SOK_FAILED;
			}
			FD_ZERO(&reading);
			FD_SET(sok, &reading);
			if (tv!=NULL) {
				memcpy(tv, &tvbak, sizeof(*tv));
			}
			do {
				ret = select(sok+1, &reading, NULL, NULL, tv);
			} while (ret==-1 && errno==EINTR);
			if (ret==-1) {
				free(msg);
				return SOK_FAILED;
			} else if (ret==0) {
				free(msg);
				return SOK_TIMEOUT;
			}
		} else if (ret==0) {
			return SOK_FAILED; /* EOF */
		} else {			
			if ('\n'==*(msg+numchars)) {
				/* telnet likes to end with '\r\n' */
				if ((numchars>0) && ('\r'==*(msg+numchars-1))) {
					numchars--;
				}
				*(msg+numchars) = '\0';
				return SOK_SUCCESS;
			}
			numchars++;
		}
	}
	/* string ending in \r possible */
	*(msg+numchars) = '\0';
	return SOK_TOOLONG;
}

char *makepeername(int sok)
{
	char *name;
	struct sockaddr_in peerstruct;
	int peerlen = sizeof(peerstruct);

	memset(peerstruct.sin_zero, 0, sizeof(peerstruct.sin_zero));
	if ((bool)getpeername(sok, (struct sockaddr *)&peerstruct, &peerlen)) {
		notice_err("getpeername() failed");
		name = NULL;
	} else {
		if (alloc_copy(&name, inet_ntoa(peerstruct.sin_addr))) {
			notice("getting peername failed\n");
			name = NULL;
		}
	}
	return name;
}
